/*------------------------------------------------------------------------
 *
 * File:        preamp.c
 *
 * Author:		Suzy Jackson <suzyj@littlefishbicycles.com>
 *
 * Revision:	0.1
 *
 * Commenced:	16 April 2006.
 *
 * Last edit:	2 May 2006.
 *
 * Purpose:     Preamp controller.
 *
 * Inputs:      Rotary encoder switch (volume control).
 *
 *              Eight source select pushbuttons.
 *
 *              Power pushbutton.
 *
 *              IR receiver serial stream.
 *
 * Outputs:     VFD for source display and volume level display.
 *
 *              Serial commands to Burr Brown PGA2310 stereo volume
 *              control.
 *
 *              Three address lines for video switch.
 *
 *              Eight outputs for audio source select relays.
 *
 *              Power enable line to allow preamp analogue board, 
 *              video switch board, and power amps to be shut down
 *              when not in use.
 *
 * Description: This is the firmware for a stereo preamplifier.
 *              It allows the user to select between eight stereo audio
 *              and S-video sources, using relays to select the audio,
 *              and video muxes to select the video.
 *              The preamp uses a Burr Brown PGA2310 stereo volume
 *              control chip to allow volume control from an infra-red
 *              remote controller, without mucking around with motorised
 *              pots, and indeed without mucking around with pots at all.
 *              The preamp tells the user what source is selected, as
 *              well as the current volume setting using a Noritake-Itron
 *              GU140X16G graphical Vacuum Fluorescent Display (VFD).  The
 *              display pretends it's a character LCD, to simplify control.
 *              The preamp also takes inputs from an IR receiver chip. 
 *
 *------------------------------------------------------------------------*/

/*------------------------------------------------------------------------
 *
 *                          I/O Map
 *                          =======
 *
 * Port Bit I/O Pin     to Where        Description
 * ---- --- --- ---     --------        -----------
 *
 *   A   0   i  3       Unused
 *   A   1   i  4       Unused
 *   A   2   i  5       Unused
 *
 *   A   3   o  6       PWR_EN	        Low enables power supply 
 *                                      for VFD, Preamp, Video Switch, 
 *                                      Power amp, Volume rotary encoder.
 *                                      When high, only pushbuttons and 
 *                                      IR receiver will work.
 *
 *   A   4   i  7       ENC_A           Rotary encoder A input.
 *
 *   A   5   i  8       ENC_B           Rotary encoder B input.
 *
 *   B   0   i  36      IR              Serial stream from IR receiver. 
 *
 *   B   1   o  37      PB_EN           Low gates source select 
 *                                      pushbuttons onto port D.
 *
 *   B   2   o  38      RELAY_CLK       Low to high transition clocks
 *                                      Port D onto relay driver.
 *
 *   B   3   i  39      Unused
 *
 *   B   4   i  41      PWRPB           Input from Power pushbutton.
 *                                      Low = button pressed.
 *
 *   B   5   i  42      Unused
 *   B   6   i  43      Unused
 *   B   7   i  44      Unused
 *
 *   C   0   o  16      VOL_CS          Low enables PGA2310 
 *                                      Volume control.
 * 
 *   C   1   o  18      VOL_MUTE        Low mutes PGA2310.
 * 
 *   C   2   o  19      VFD_RS          Register select for VFD.
 * 
 *   C   3   o  20      VFD_RW          VFD Read/-Write line.
 * 
 *   C   4   o  25      VFD_EN          VFD E Strobe.
 * 
 *   C   5   o  26      VID_A0          Video Mux A0.
 * 
 *   C   6   o  27      VID_A1          Video Mux A1.
 * 
 *   C   7   o  29      VID_A2          Video Mux A2.
 * 
 *   D   0   io 21      D0              Data 0.  Used by VFD, Relay
 *                      VOL_D           driver, and pushbuttons.
 *                                      High enables source 1 when
 *                                      RELAY_clk clocked.  Low 
 *                                      when PB_en low means source 1
 *                                      pushbutton pressed.
 *                                      PGA2310 data when VOL_cs low. 
 *
 *   D   1   io 22      D1              Data 1.  Used by VFD, Relay
 *                      VOL_CLK         driver, and pushbuttons.
 *                                      High enables source 2 when
 *                                      RELAY_clk clocked.  Low 
 *                                      when PB_en low means source 2
 *                                      pushbutton pressed.
 *                                      PGA2310 clock when VOL_cs low. 
 *
 *   D   2   io 23      D2              Data 2.  Used by VFD, Relay
 *                                      driver, and pushbuttons.
 *                                      High enables source 3 when
 *                                      RELAY_clk clocked.  Low 
 *                                      when PB_en low means source 3
 *                                      pushbutton pressed.
 *
 *   D   3   io 24      D3              Data 3.  Used by VFD, Relay
 *                                      driver, and pushbuttons.
 *                                      High enables source 4 when
 *                                      RELAY_clk clocked.  Low 
 *                                      when PB_en low means source 4
 *                                      pushbutton pressed.
 *
 *   D   4   io 30      D4              Data 4.  Used by VFD, Relay
 *                                      driver, and pushbuttons.
 *                                      High enables source 5 when
 *                                      RELAY_clk clocked.  Low 
 *                                      when PB_en low means source 5
 *                                      pushbutton pressed.
 *
 *   D   5   io 31      D5              Data 5.  Used by VFD, Relay
 *                                      driver, and pushbuttons.
 *                                      High enables source 6 when
 *                                      RELAY_clk clocked.  Low 
 *                                      when PB_en low means source 6
 *                                      pushbutton pressed.
 *
 *   D   6   io 32      D6              Data 6.  Used by VFD, Relay
 *                                      driver, and pushbuttons.
 *                                      High enables source 7 when
 *                                      RELAY_clk clocked.  Low 
 *                                      when PB_en low means source 7
 *                                      pushbutton pressed.
 *
 *   D   7   io 33      D7              Data 7.  Used by VFD, Relay
 *                                      driver, and pushbuttons.
 *                                      High enables source 8 when
 *                                      RELAY_clk clocked.  Low 
 *                                      when PB_en low means source 8
 *                                      pushbutton pressed.
 *
 *   E   0   i  9       Unused
 *   E   1   i  10      Unused
 *   E   2   i  11      Unused
 *
 *------------------------------------------------------------------------*/

#include <pic.h>
#include "hardware.h"
#include "vfd.h"
#include "delay.h"
#include "preamp.h"

/*------------------------------------------------------------------------
 *
 * VOLUME SET - Drives serial data lines on PGA2310 volume control.
 *
 *------------------------------------------------------------------------*/

void
volume_set(unsigned char right, unsigned char left)
{
	unsigned char i, b;
	VOL_CS = 0;	// Enable the volume control
	b=right;
	for (i=8;i>0;i--) {	// Clock out right data
		unsigned char x= 1 << (i-1);
		if (b>=x) {
			VOL_D=1;
			b-=x;
		}
		else {
			VOL_D=0;
		}
		VOL_CLK = 0;
		VOL_CLK = 1;
	}
	b=left;
	for (i=8;i>0;i--) {	// Clock out right data
		unsigned char x= 1 << (i-1);
		if (b>=x) {
			VOL_D=1;
			b-=x;
		}
		else {
			VOL_D=0;
		}
		VOL_CLK = 0;
		VOL_CLK = 1;
	}
	VOL_CS = 1;		// Disable the volume control
}	

/*------------------------------------------------------------------------
 *
 * ENCODER - Checks rotary encoder and returns the following:
 *
 *               0:  no movement
 *              -1:  ccw rotation
 *               1:  cw rotation
 *
 *------------------------------------------------------------------------*/

signed char
encoder(void)
{
	static unsigned char a, b;
	if (!a & !b) {
		if (ENC_A) {
			a = ENC_A;
			return (-1);
		}
		if (ENC_B) {
			b = ENC_B;
			return (1);
		}
	}
	if (a & !b) {
		if (!ENC_A) {
			a = ENC_A;
			return (1);
		}
		if (ENC_B) {
			b = ENC_B;
			return (-1);
		}
	}
	if (a & b) {
		if (!ENC_A) {
			a = ENC_A;
			return (-1);
		}
		if (!ENC_B) {
			b = ENC_B;
			return (1);
		}
	}
	if (!a & b) {
		if (ENC_A) {
			a = ENC_A;
			return (1);
		}
		if (!ENC_B) {
			b = ENC_B;
			return (-1);
		}
	}
	return (0);
}

/*------------------------------------------------------------------------
 *
 * RC5_RECEIVE - Checks IR receiver and returns the following:
 *
 *               0:  no meaningful IR data present
 *              !0:  14 bit RC5 packet
 *
 *------------------------------------------------------------------------*/

unsigned int
rc5_receive(void)
{
	unsigned int message;
	unsigned char i, cnt, IR_old;
	message = 0;
	if (!IR) {	// in middle of start bit
		for (i=0;i<14;i++) {
			message = message << 1 + IR;
			cnt = 6;	// wait 1500ms (just before middle of next bit at 1778ms)
			do {
				DelayUs(250);
			} while (--cnt);
			IR_old = IR;
			cnt = 50;
			do {
				DelayUs(10);
			} while (IR_old == IR & --cnt);
			if (!cnt) return (0);	// cnt = 0 indicates timeout - return 0
		}
	}
	return (message);
}

/*------------------------------------------------------------------------
 *
 * CHECK_SOURCE_PB - Checks source select pushbuttons and returns the following:
 *
 *               0xff:  no pushbutton pressed
 *               0x00:  source 1 pushbutton pressed
 *               ...
 *               0x07:  source 8 pushbutton pressed
 *
 *------------------------------------------------------------------------*/

unsigned char
check_source_pb(void)
{
	unsigned char source, i;
	TRISD = 0xff;	// set data port to input 
	PB_EN = 0;		// enable pushbutton buffer
	source = DATA;
	PB_EN = 1;		// disable pushbutton buffer
	TRISD = 0x00;	// set data port to output 
	if (source == 0x00) return (0xff);	// no button pressed
	DelayMs (10);	// debounce delay
	TRISD = 0xff;	// set data port to input 
	PB_EN = 0;		// enable pushbutton buffer
	source = DATA;
	PB_EN = 1;		// disable pushbutton buffer
	TRISD = 0x00;	// set data port to output 
	switch (source) {
		case 0b00000001:
			return (0x00);
			break;
		case 0b00000010:
			return (0x01);
			break;
		case 0b00000100:
			return (0x02);
			break;
		case 0b00001000:
			return (0x03);
			break;
		case 0b00010000:
			return (0x04);
			break;
		case 0b00100000:
			return (0x05);
			break;
		case 0b01000000:
			return (0x06);
			break;
		case 0b10000000:
			return (0x07);
			break;
		default:
			return (0xff);
			break;
	}
	return (0xff);	// no button pressed
}

/*------------------------------------------------------------------------
 *
 * RELAY_SET - Outputs data to the relay driver.  Inputs are:
 *
 *               0x00:  source 1 selected
 *               ...
 *               0x07:  source 8 selected
 *
 *------------------------------------------------------------------------*/

void
relay_set (unsigned char source)
{
	switch (source) {
		case 0:
			DATA = 0b00000001;
			break;
		case 1:
			DATA = 0b00000010;
			break;
		case 2:
			DATA = 0b00000100;
			break;
		case 3:
			DATA = 0b00001000;
			break;
		case 4:
			DATA = 0b00010000;
			break;
		case 5:
			DATA = 0b00100000;
			break;
		case 6:
			DATA = 0b01000000;
			break;
		default:
			DATA = 0b10000000;
			break;
	}
	RELAY_CLK = 0;
	RELAY_CLK = 1;
	RELAY_CLK = 0;
}

/*------------------------------------------------------------------------
 *
 * VIDEO_SET - Outputs data to the video select board.  Inputs are:
 *
 *               0x00:  source 1 selected
 *               ...
 *               0x07:  source 8 selected
 *
 *------------------------------------------------------------------------*/

void
video_set (unsigned char source)
{
	VID_A0 = source & 0b00000001;
	VID_A1 = source & 0b00000010;
	VID_A2 = source & 0b00000100;

}

/*------------------------------------------------------------------------
 *
 * UPDATE_LEGENDS - displays the dB symbol and decimal point on the VFD.
 *
 *------------------------------------------------------------------------*/

void
update_legends(void)
{
	const char db_symbol [20] = {            
		0b00000001,	0b00000001,	0b00000001,	0b00000001,	0b00011101,
		0b00100011, 0b00100001,	0b00100001,	0b00100011,	0b00011101,	0b00111110,
		0b00100001, 0b00100001, 0b00100001, 0b00111110, 0b00100001,
		0b00100001, 0b00100001, 0b00100001, 0b00111110};
	unsigned char i;
	vfd_area(110, 7, 111, 8, 'F');	// print decimal point           //  300us
	vfd_area(124, 1, 139, 10, 'h');                                  //  300us
	for (i=0;i<20;i++) vfd_write_data(db_symbol[i]);	// print dB symbol   //  880us
}

/*------------------------------------------------------------------------
 *
 * UPDATE_SOURCE_DISPLAY - Displays thesource on the VFD.
 *
 *           Source selection:
 *               0x00:  "Tuner"
 *               0x01:  "TV"
 *               0x02:  "Phono"
 *               0x03:  "DVD"
 *               0x04:  "CD"
 *               0x05:  "VHS"
 *               0x06:  "Tape"
 *               0x07:  "HTPC"
 *
 *------------------------------------------------------------------------*/

void
update_source_display(unsigned char source)
{
	vfd_goto(0,15);
	switch (source) {
		case 0:
			vfd_puts("Tuner");
			break;
		case 1:
			vfd_puts("TV   ");
			break;
		case 2:
			vfd_puts("Phono");
			break;
		case 3:
			vfd_puts("DVD  ");
			break;
		case 4:
			vfd_puts("CD   ");
			break;
		case 5:
			vfd_puts("VHS  ");
			break;
		case 6:
			vfd_puts("Tape ");
			break;
		default:
			vfd_puts("HTPC ");
			break;
	}
}

/*------------------------------------------------------------------------
 *
 * UPDATE_VOLUME_DISPLAY - Displays the volume on the VFD.
 *
 *           Volume encoding:
 *               0x00:  "-96.0dB"
 *               0x01:  "-95.5dB"
 *               ...
 *               0xc0:  "+00.0dB"
 *               ...
 *               0xfe:  "+31.0dB"
 *               0xff:  "+31.5dB"
 *
 *------------------------------------------------------------------------*/

void
update_volume_display(unsigned char volume)
{
	const char plus_symbol [6] = {            
		0b00001100, 0b00001100, 0b00111111, 0b00111111, 0b00001100, 0b00001100};
	const char minus_symbol [6] = {            
		0b00000000, 0b00000000, 0b00111111, 0b00111111, 0b00000000, 0b00000000};
	unsigned char i;
	vfd_area(76, 5, 83, 10, 'h');
	if (volume>=192) {
		for (i=0;i<6;i++) vfd_write_data(plus_symbol[i]);	// print + symbol
		vfd_goto(86,15);
		vfd_write_data((volume-192)/20+'0');
		vfd_write_data((volume-192)/2%10+'0');
	}
	else {
		for (i=0;i<6;i++) vfd_write_data(minus_symbol[i]);	// print - symbol
		vfd_goto(86,15);
		vfd_write_data((192-volume)/20+'0');
		vfd_write_data((192-volume)/2%10+'0');
	}
	vfd_goto(114,15);
	if (volume%2) {
		vfd_write_data('5');
	}
	else {
		vfd_write_data('0');
	}
}

/*------------------------------------------------------------------------
 *
 * UPDATE_DISPLAY - displays the whole display.
 *
 *------------------------------------------------------------------------*/

void
update_display(unsigned char source, unsigned char volume)
{
	vfd_clear();
	vfd_font('c');	// Set 10x14 font
	vfd_font('2');	// Set 1 pixel inter-char spacing
	update_legends();
	update_source_display(source);
	update_volume_display(volume);
}

/*------------------------------------------------------------------------
 *
 * MAIN - main code.
 *
 *------------------------------------------------------------------------*/

void
main(void)
{
	unsigned char volume = 0;
	unsigned char source = 0;
	signed char enc;
	unsigned char temp;
	unsigned int ir_packet;
	ADCON1 = 0x07;			// disable ADC
	TRISA = 0b11110111;		// set directions for port a
	TRISB = 0b11111001;		// set directions for port b
	TRISC = 0b00000000;		// set directions for port c
	TRISD = 0b00000000;		// set directions for port d
	PWR_EN = 1;	// start with everything switched off
	enc = encoder();	// dummy encoder read to initialise statics
	for(;;) {
		if (!PWR_EN) {
			if ((enc = encoder()) != 0) {
				if (volume > 0 & enc < 0) {
					volume--;
				}
				if (volume < 255 & enc > 0) {
					volume++;
				}
				update_volume_display(volume);
				enc = encoder();	// dummy encoder read to account for slow display update
				volume_set(volume, volume);
				VOL_MUTE = 1;
			}
			temp = check_source_pb();
			if (temp <= 0x07) {
				source = temp;
				update_source_display(source);
				relay_set(source);
				video_set(source);
			}
			if (PWR_PB == 0) {	// power button pressed
				DelayMs(10);	// debounce delay
				if (PWR_PB == 0) {
					VOL_MUTE = 0;
					PWR_EN = 1;	// shut down unnecessary stuff
					do {
						DelayMs(10);
					} while (PWR_PB == 0);  // wait for button to be released
					DelayMs(10);
				}
			}
		}
		else {
			if (PWR_PB == 0) {	// power button pressed
				DelayMs(10);	// debounce delay
				if (PWR_PB == 0) {
					PWR_EN = 0;	// power stuff up
					VOL_MUTE = 1;
					DelayMs(10);
					vfd_init();
					update_display(source, volume);
					volume_set(volume, volume);
					relay_set(source);
					video_set(source);
					do {
						DelayMs(10);
					} while (PWR_PB == 0);  // wait for button to be released
					DelayMs(10);
					enc = encoder();	// dummy encoder read to initialise statics
				}
			}
		}
	}
}
	
